


/*
  SOLAR INVERTER
  Program stridace pro fotovoltaicke panely a ohrev bojleru.
  Prevadi SS proud z FV panelu na stridavy obdelnikovy s promennou stridou.
  Vyhledava bod nejvyssiho vykonu fv panelu, zobrazeni hodnot na I2C LCD
  pres expander PCF8574
  mereni napeti a proudu, tlacitka pro rucni zmenu stridy menice
  rucni a automaticky rezim, vypocet vyroby EE, pri zvyseni vyroby o 0,1kWh
  ulozi vyrobu do EEprom, komunikace po RS485 s EmonHUBem prenos hodnot na EmonCMS.

  Verze 4.3a
  kompilace Arduino 1.6.7
  Tomas Nevrela; tnweb.tode.cz
  1.2.2017
	Zmeny:
  4.2  - opravena chyba nacitani vyroby

  4.3  - opraveny adresy v eeprom po 90000 zapisech
	     - automaticka inicializace Eeprom po prvnim nahrani programu

  4.3a - zvysena zmena pwm na -3 a +4, prodleva 150ms pred novym merenim
			 - zobrazeni stridy na LCD v aut. rezimu
			 - zvysen cas prepnuti z ruc. do aut. rezimu na 10s


*/
#include <SoftEasyTransfer.h>
#include <SoftwareSerial.h>
#include <EEPROM.h>
#include "EEPROMAnything.h"
#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include <avr/wdt.h> //nechat jen pri progamovani pres ISP
SoftwareSerial mySerial(2, 4); //Arduino RX - RS485 RO, Arduino TX - RS485 DI
#define TXenableRS485 3 //RE + DE
#define tlacPlus 5 //tlacitko plus strida
#define tlacMinus 6 //tlacitko minus strida
#define pinA A2 //pin mereni proudu fv
#define pinV A1 //pin mereni napeti fv
#define R1 470000.0 //odpory delice napeti
#define R2 5600.0
#define maxStrida 247 //maximalni strida (nemenit!)
#define frekvence 50 //frekvence PWM vystupu (nemenit!)
#define LEDpin 13  // signalizacni Emon LED
const int nodeID = 12; // cislo Emon nodu
int interval_LCD = 800; //cas obnoveni LCD
int rucTimeExit = 10000; //cas prepnuti z rucniho zpet do aut. rezimu v ms
unsigned long currentMillis = 0;
unsigned long previousMillis = 0;
unsigned long previousVA = 0;
unsigned long rucTime;
unsigned long vyrobaTime = 0;
unsigned long vyroba = 0;
unsigned long lastvyroba;
float whInc2;
boolean zapis;
int strida = 10;
long topv;
int amp, nap, vcc;
float vPow;
int napeti, proud, vykon;
int vykon_prev;
int smer = 1; //smer zmeny stridy 1 = zvysovani 2 = snizovani
boolean rucne = false; //rezim rizeni
int addr_vyroba = 20;
int addr_suma_write = 25;
unsigned long SumaWrite; // aktualni pocet zapisu do EEPROM
unsigned long PreviousWrite;// predchozi hodnota zapisu do EEPROM ++90000
unsigned long FirstRun;// hodnota kontroly prvniho spusteni programu


// struktura pro komunikaci s emonHUB
struct PayloadTX {
  byte nodeID, address, command, func;
  int data1, data2, data3, data4, data5, data6, data7, data8;
  long data9, data10, data11, data12;
};
PayloadTX emontx;   // vytvoreni instance
SoftEasyTransfer ET;
// Nastav pro LCD piny PCF8574T a I2C addresu 0x27(PCF8574AT 0x3f)
//                    addr, en,rw,rs,d4,d5,d6,d7,bl,blpol
LiquidCrystal_I2C lcd(0x27, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE );



//=================================Setup========================================
//==============================================================================
void setup() {
  wdt_enable(WDTO_4S);// aktivace watchdogu(nechat jen pri progamovani pres ISP)
  wdt_reset(); // vynulovani watchdogu (nechat jen pri progamovani pres ISP)
  //==============================================================================
  // Zapis vychozich hodnot do Eeprom pri prvnim spusteni programu.

  //nacte kontrolni hodnotu z konce Eeprom
  EEPROM_readAnything(1020, FirstRun);

  //pokud hodnota nesouhlasi provede se prvotni inicializace Eeprom
  if (!(FirstRun == 0x710812AA))
  {
    EEPROM_writeAnything(2, addr_vyroba);
    EEPROM_writeAnything(6, addr_suma_write);
    EEPROM_writeAnything(8, PreviousWrite);
    EEPROM_writeAnything(addr_vyroba, vyroba);
    EEPROM_writeAnything(addr_suma_write, SumaWrite);
    //zapise kontrolni hodnotu o provedene prvotni inicializaci
    FirstRun = 0x710812AA;
    EEPROM_writeAnything(1020, FirstRun);
  }
  // =============================================================================
  // odkomentovat pokud je potreba ulozit do Eeprom predchozi vyrobu,
  // nahrat program, spustit pak znovu zakomentovat, nahrat program a spustit
  // pozn. Pri programovani pres ISP a Arduino IDE se EEPROM maze, aby se tak
  // nestalo je treba nastavit fusebit EESAVE na 0. Normalnim nahranim pres bootloader
  // se EEPROM zachova.

  //vyroba = 2204500; //vyroba ve Wh (napr.1457600 = 1457,600kWh)
  //EEPROM_writeAnything(addr_vyroba,vyroba);

  //==============================================================================

  EEPROM_readAnything(2, addr_vyroba); //adresa pro nacteni sumy vyroby
  EEPROM_readAnything(6, addr_suma_write); //adresa pro nacteni sumy zapisu do EEPROM
  EEPROM_readAnything(8, PreviousWrite); //predchozi hodnota poctu zapisu do EEPROM
  EEPROM_readAnything(addr_vyroba, vyroba); //nacteni vyroby z aktualni adresy
  EEPROM_readAnything(addr_suma_write, SumaWrite);// nacteni sumy zapisu z akt. adresy
  lastvyroba = vyroba;

  pinMode(TXenableRS485, OUTPUT);
  digitalWrite(TXenableRS485, LOW);
  pinMode(LEDpin, OUTPUT);
  pinMode(tlacPlus, INPUT_PULLUP);
  pinMode(tlacMinus, INPUT_PULLUP);
  //Serial.begin(9600);
  // softwareSerial pro rs485
  mySerial.begin(9600);
  ET.begin(details(emontx), &mySerial);// nastaveni EasyTransfer

  lcd.begin(16, 2); // pocet znaku, pocet radku
  lcd.backlight(); // zapni podsvetleni (noBacklight -vypni)
  lcd.setCursor(0, 0);
  lcd.print(" SOLAR INVERTER ");
  lcd.setCursor(0, 1);
  lcd.print("v4.3a   1.2.2017");
  delay(3000);
  lcd.clear();
  wdt_reset();//nechat jen pri progamovani pres ISP
}

//==============================Hlavni smycka===================================
//==============================================================================
void loop() {
  mereni();
  rizeni();
  set_PWM(strida);
  delay(10);//
  zobrazeni();
  if (!rucne) delay(150); // zpozdeni pokud neni rucni rezim

  if (ET.receiveData())
  {
    // kontrola jestli je pozadavek na data z emonHUBu
    if (emontx.nodeID == 15 && emontx.address == 12 && emontx.command == 5)
    {
      // naplneni struktury
      emontx.data1 = napeti;
      emontx.data2 = proud;
      emontx.data3 = vykon;
      emontx.data9 = vyroba; //vyroba ve Wh
      emontx.nodeID = nodeID; // ID solar inventoru
      emontx.command = 6; // odpoved na pozadavek
      emontx.address = 15; // data pro emonHUB
      digitalWrite(LEDpin, HIGH);// activity LED
      delay(15);
      digitalWrite(TXenableRS485, HIGH);//prepni prevodnik RS485 na vysilani
      ET.sendData(); // odesli data na emonHUB
      digitalWrite(TXenableRS485, LOW);
      delay(20);
      digitalWrite(LEDpin, LOW); // activity LED
    }
  }

  // Zapis do EEPROM se provede pokud se vyroba zvysi o 100Wh
  //  posun adresy po 90000 zapisech
  if (zapis)
  {
    EEPROM_writeAnything(addr_vyroba, vyroba); //uloz vyrobu
    SumaWrite++;
    zapis = false;
    EEPROM_writeAnything(addr_suma_write, SumaWrite); //uloz celkovy pocet zapisu
    // pokud pocet zapisu do jedne bunky presahl 90000 posun se na dalsi
    if ((SumaWrite - PreviousWrite) >= 90000)
    {
      addr_vyroba += 10; // posun adresy pro zapis o 10
      addr_suma_write += 10;
      // zapis aktualnich hodnot do EEPROM
      EEPROM_writeAnything(2, addr_vyroba); //uloz akt. adresu vyroby
      EEPROM_writeAnything(6, addr_suma_write); //uloz akt. adresu sumy zapisu
      PreviousWrite = SumaWrite;
      EEPROM_writeAnything(8, PreviousWrite);
      EEPROM_writeAnything(addr_vyroba, vyroba);
      EEPROM_writeAnything(addr_suma_write, SumaWrite);
    }
  }

  //=============================Obsluha tlacitek==================================

  if (digitalRead(tlacPlus) == 0) {
    strida++;
    rucne = true; //po stisku tlacitka rezim zmeny stridy rucne
    rucTime = millis();
    delay(15);
  }
  if (strida > maxStrida) strida = maxStrida;

  if (digitalRead(tlacMinus) == 0) {
    strida--;
    rucne = true; //po stisku tlacitka rezim zmeny stridy rucne
    rucTime = millis();
    delay(15);
  }
  if (strida < 0) strida = 0;
}

//===============================Nastaveni PWM==================================
//==============================================================================
//Nastavi casovace PWM, max. hodnota OCR1A=1235, OCR1B=1265 pro dostatecny dead
//time (120us)
//vystup na pinech D9 a D10 casove posunute o 10ms
void set_PWM(int dutyCycle)
{
  if (dutyCycle > maxStrida) dutyCycle = maxStrida;
  else if (dutyCycle < 0) dutyCycle = 0;
  cli();
  TCCR1B = _BV(WGM13) | _BV(CS11) | _BV(CS10);// rezim 8,clock/64

  int topv = (F_CPU / (frekvence * 2 * 64));//16000000/(50*2*64)=2500
  ICR1 = topv;
  OCR1A = (dutyCycle * 5); //od 0 do 1235
  OCR1B = ( topv - (dutyCycle * 5)); //od 2500 do 1265

  DDRB |= _BV(PORTB1) | (_BV(PORTB2)); //D9(PB1), D10(PB2) nastav na vystup
  TCCR1A = _BV(COM1A1) | _BV(COM1B1) | _BV(COM1B0); //rezim vyvodu OC1A, OC1B
  sei();
}
//===============================Mereni=========================================
//==============================================================================
//mereni proudu a napeti, vypocet vykonu a vyroby, prevod potrebnych hodnot do int
void mereni()
{
  int cnt = 200; //pocet vzorku mereni
  unsigned long totA = 0;
  unsigned long totV = 0;
  float vyk, Avolt;
  wdt_reset();//nechat jen pri progamovani pres ISP
  vcc = readVcc();// zmereni napajeciho napeti 5V
  vPow = vcc / 1000.0;

  for (int i = 0; i < cnt; i++)
  {
    totA = totA + analogRead(pinA); //zmer proud
    totV = totV + analogRead(pinV); //zmer napeti
  }
  int amp = totA / cnt; // prumer ze vzorku
  int nap = totV / cnt;
  // vypocet proudu
  Avolt = amp * (vcc / 1023.0);
  proud = (Avolt - (vcc / 2)); // 0856 = 8,56A
  if (proud <= 7) proud = 0;//kalibrace 0
  float a = proud;
  float v = nap * (vPow / 1023.0);
  float v2 = v / (R2 / (R1 + R2)); //vypocet napeti podle delice
  napeti = (int)v2;
  vyk = (v2 * a) / 100; //vypocet vykonu
  vykon = (int)vyk; //okamzity vykon


  //=====================Vypocet vyroby Wh====================================
  unsigned long lvyrobaTime = vyrobaTime;
  vyrobaTime = millis();
  float whInc = vykon * ((vyrobaTime - lvyrobaTime) / 3600000.0);
  whInc2 = whInc2 + whInc;
  if (whInc2 > 1.0)
  {
    whInc2 = whInc2 - 1.0;
    vyroba++;
  }
  if ((vyroba - lastvyroba) >= 100)
  {
    zapis = true;
    lastvyroba = vyroba;
  }

}
//=======================Zmereni napeti 5V====================================
long readVcc()
{
  long result;
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  delay(2);
  ADCSRA |= _BV(ADSC);
  while (bit_is_set(ADCSRA, ADSC));
  result = ADCL;
  result |= ADCH << 8;
  result = 1126400L / result;
  return result;
}

//==========================Rizeni vykonu=======================================
//==============================================================================
//hledani max. vykonu fotovoltaickych panelu
void rizeni()
{
  currentMillis = millis();
  if ((unsigned long)(currentMillis > (rucTime + rucTimeExit))) rucne = false;
  if (rucne == false)
  {

    if (vykon == 0) {
      strida = 5;  //pri nulovem vykonu nastav zakladni stridu a smer
      smer = 1; //zvysovani
    }
    //smer nahoru
    switch (smer) {
      case 1:
        {
          if (vykon >= vykon_prev) {
            strida += 4; //pokud se vykon zvysuje zvys stridu
            if (strida > maxStrida) {
              strida = maxStrida;
              smer = 2;
            }
          }

          else smer = 2; //pokud je vykon nizsi zmen smer
          break;
        }
      //smer dolu
      case 2:
        {
          if (vykon >= vykon_prev && vykon > 0) {
            strida -= 3; //pokud se vykon zvysuje sniz stridu
            if (strida < 2) {
              strida = 2;
              smer = 1;
            }
          }
          else smer = 1; //pokud je vykon nizsi zmen smer
          break;
        }
    }
    vykon_prev = vykon;
  }
}

//==============================Zobrazeni na LCD================================
//==============================================================================
void zobrazeni()
{
  currentMillis = millis();
  if ((unsigned long)(currentMillis - previousMillis) >= interval_LCD) {

    lcd.setCursor(0, 0);// sloupec, radek
    if (napeti < 100 && napeti > 9) lcd.print(" ");
    if (napeti < 10) lcd.print("  ");
    lcd.print(napeti);
    lcd.print("V ");

    lcd.setCursor(5, 0);
    if (proud < 1000) lcd.print(" ");
    lcd.print(proud / 100);
    lcd.print(".");
    lcd.print((proud % 100) / 10);
    lcd.print("A ");

    lcd.setCursor(11, 0);
    if (vykon < 1000 && vykon > 99) lcd.print(" ");
    if (vykon < 100 && vykon > 9) lcd.print("  ");
    if (vykon < 10) lcd.print("   ");
    lcd.print(vykon);
    lcd.print("W");

    if (rucne)
    {
      lcd.setCursor(0, 1);
      lcd.print("strida:");
      if ((int)(strida / 2.5) < 10) lcd.print(" ");
      lcd.print((int)(strida / 2.5));
      lcd.print("%      ");

    }
    else {
      lcd.setCursor(0, 1);
      lcd.print(vyroba / 1000);
      lcd.print(".");
      if (((vyroba % 1000) / 10) < 10) lcd.print("0");
      lcd.print((vyroba % 1000) / 10);
      lcd.print("kWh     ");
      lcd.setCursor(13, 1);
      if ((int)(strida / 2.5) < 10) lcd.print(" ");
      lcd.print((int)(strida / 2.5));
      lcd.print("%");

    }
    previousMillis = currentMillis;
  }
}


